/*
** scan_api.c - written in milano by vesely on 9 jan 2017
** clamav interface for courier-mta global filters
**
** This file is part of avfilter
**
** Copyright (c) 2017 Alessandro Vesely
** All rights reserved. see COPYING
*/
#define _GNU_SOURCE
#if defined(HAVE_CONFIG_H)
#include <config.h>
#endif

#include <dlfcn.h>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <clamav.h>
#include <syslog.h> // for LOG_XXXX defs

#include "scan_api.h"
#include "log.h"

#include <assert.h>

static int add_name(virus_name_list **l, char const *name)
{
	assert(name != NULL);

	if (name)
	{
		size_t const len = strlen(name);
		virus_name_list *n = malloc(len + sizeof(virus_name_list));
		if (n)
		{
			n->next = *l;
			memcpy(&n->name[0], name, len + 1);
			*l = n;
			return 0; // good
		}
	}

	return -1; // fail
}

void free_virus_name_list(virus_name_list *l)
{
	while (l)
	{
		virus_name_list *t = l->next;
		free(l);
		l = t;
	}
}

#define DEAD_BEEF 0xdeadbeefUL

typedef struct vir_names
{
	virus_name_list *l;
	unsigned long dead_beef;
	unsigned count, errors;
} vir_names;

static void virus_found_cb(int fd, const char *virname, void *context)
{
	if (context)
	{
		vir_names *vn = (vir_names*)context;
		if (vn->dead_beef != DEAD_BEEF)
			filelog(stderr, LOG_CRIT, "callback error: BAD CONTEXT");
		else
		{
			vn->count += 1;
			if (virname == 0)
				filelog(stderr, LOG_CRIT, "callback error: no name");
			else if (add_name(&vn->l, virname))
				++vn->errors;
		}
	}

	return;

	(void)fd; // not used
}

union clam_hack
/*
*  This was used before introducing the callback, sometimes about 0.99.2
*  see libclamav/others.h: #define SCAN_ALL  (ctx->options & CL_SCAN_ALLMATCHES)
*  and libclamav/others.c: cli_append_virus(), called by many scanners, and then
*  how libclamav/scanners.c: scan_common() copies the resulting array back, compare
*  with how clamd/scanner.c: scan_callback() frees each entry
*/
{
	char const *name;
	char const **pp;
};


int my_scandesc(int desc, virus_name_list **vnl, const struct cl_engine *engine, unsigned int scanoptions)
/*
* call scan function with callback;
* if a virus is found but callback never called, assume old version and call
* the callback before return.
*/
{
	union clam_hack vir = {NULL};
	vir_names lst;
	unsigned long int scanned;
	int cl;

	memset(&lst, 0, sizeof lst);
	lst.dead_beef = DEAD_BEEF;
	cl = cl_scandesc_callback(desc, &vir.name, &scanned, engine, scanoptions, &lst);
	if (lst.count == 0 && cl == CL_VIRUS)
	/*
	* old version, the callback was not called
	*/
	{
		unsigned count;
		if ((scanoptions & CL_SCAN_ALLMATCHES) != 0)
		{
			for (count = 0; vir.pp[count] != NULL; ++count)
				virus_found_cb(0, vir.pp[count], &lst);

			free(vir.pp);
		}
		else
			virus_found_cb(0, vir.name, &lst);
	}

	if (lst.errors == 0)
		*vnl = lst.l;
	else
	{
		filelog(stderr, LOG_CRIT,
			"callback error(s): %d out of %d calls", lst.errors, lst.count);
		free_virus_name_list(lst.l);
		*vnl = NULL;
	}

	return cl;
}


void init_scan_api(struct cl_engine *clamav, int verbose)
{
	/*
	this function is only called once on loading

	for newer clamav versions:
	clcb_virus_found is defined in libclamav/clamav.h
	cl_engine_set_clcb_virus_found is defined in libclamav/others.c

	here we provide alternative definitions, so as to be compatible with older
	versions of libclamav, and set the callback if at all possible
	*/

	typedef void (*my_clcb_virus_found)(int fd, const char *virname, void *context);
	void (*engine_set_clcb_virus_found)(struct cl_engine *, my_clcb_virus_found);
	char *error;

	dlerror();
	*(void**)(&engine_set_clcb_virus_found) =
		dlsym(RTLD_DEFAULT, "cl_engine_set_clcb_virus_found");

	if ((error = dlerror()) != NULL)
	{
		if (verbose >= 1)
			filelog(stderr, LOG_INFO,
				"assuming old ClamAV version (no callback): error was \"%s\"",
					error);
	}
	else
	{
		if (verbose >= 1)
			filelog(stderr, LOG_INFO,
				"assuming new ClamAV version (with callback)");
		(*engine_set_clcb_virus_found)(clamav, virus_found_cb);
	}
}

